# Generated by Django 4.2.10 on 2024-05-05 12:34

from typing import List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0040_alter_processlock_name"),
        ("database", "0040_merge_20240504_1010"),
    ]

    operations: List[str] = []
